/* zethereal.cpp
 *
 * ZEthereal -- Ethereal for the Sharp Zaurus
 * Copyright (C) 2003  Philippe Biondi <biondi@cartel-securite.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <qpushbutton.h>
#include <qlistview.h>
#include <qlineedit.h>
#include <qtextview.h> 
#include <qtabwidget.h> 
#include <qcheckbox.h> 
#include <qcombobox.h>
#include <iostream.h> 
#include <string.h> 
#include <qheader.h>

#include "zethereal.h"
#include "engine.h"

extern "C" {
#include <glib.h>
#include <stdio.h>
}

Zethereal::Zethereal( QWidget* parent,  const char* name, WFlags fl )
    : ZetherealMain( parent, name, fl )
{
	listSize = 0;
	dissectionSize = 0;
	dumpSize = 0;

	
	packetDissection->header()->hide();
	packetDissection->setSorting(-1);

    	connect(packetList, SIGNAL(currentChanged(QListViewItem *)),
	  	this, SLOT(dissect_packet(QListViewItem *)));

    	connect(zoomList, SIGNAL(clicked()), SLOT(expandList()));
    	connect(zoomDissection, SIGNAL(clicked()), SLOT(expandDissection()));
    	connect(zoomDump, SIGNAL(clicked()), SLOT(expandDump()));

	/*    	connect(packetDissection, SIGNAL(doubleClicked(QListViewItem *)),
	  	this, SLOT(expandDissection(QListViewItem *)));
    	connect(packetDump, SIGNAL(doubleClicked( int para, int pos )),
	this, SLOT(expandDump()));*/

	connect(debugButton, SIGNAL(clicked()), SLOT(debug()));
	connect(openButton, SIGNAL(clicked()), SLOT(open()));
	connect(startButton, SIGNAL(clicked()), SLOT(capture()));
	connect(stopButton, SIGNAL(clicked()), SLOT(stopCapture()));

}


Zethereal::~Zethereal()
{
}


void Zethereal::init_engine(Engine *eng)
{
	GList *l, *e;
       
	engine = eng;
	l = engine->get_interfaces();
	if (l) {
		interface->removeItem(0);
		for  (e = g_list_first(l); e != NULL; e = g_list_next(e)) {
			interface->insertItem(QString((char *)e->data));}
		interface->setEnabled( TRUE );
		interface->setCurrentItem(0);
		startButton->setEnabled( TRUE );
		engine->free_interfaces(l);
	}
	
	update_engine();

}


void Zethereal::update_engine(void)
{
	
	engine->set_count(count->text().toInt());
	engine->set_filter((const char *)filterString->text());
	engine->set_rfilter((const char *)rfilterString->text());
	engine->set_interface((const char *)interface->currentText());
	engine->set_filename((const char *)filename->text());
	engine->set_promisc((int)promisc->isChecked());
	engine->set_resolv_flags((int)DNSResolv->isChecked(),
				 (int)MACResolv->isChecked(),
				 (int)PortResolv->isChecked());
}

void Zethereal::clearCapture(void)
{
	packetList->clear();
	packetDissection->clear();
	packetDump->setText(QString());
	mpktvec.clear();
	myno=0;
}


bool Zethereal::ensureNotLast(QButton *but)
{
	if (zoomList->isOn() && zoomDissection->isOn() && zoomDump->isOn()) {
		but->toggle();
		return FALSE;
	}
	return TRUE;
}


void Zethereal::expandList()
{
	if (! ensureNotLast(zoomList)) return;
	QSize s = packetList->maximumSize();
	if (listSize) {
		s.setHeight(listSize);
		listSize = 0;
		packetList->setMaximumSize(s);
	}
	else {
		listSize = s.height();
		s.setHeight(0);
		packetList->setMaximumSize(s);
	}
}

void Zethereal::expandDissection() 
{
	if (! ensureNotLast(zoomDissection)) return;
	QSize s = packetDissection->maximumSize();
	if (dissectionSize) {
		s.setHeight(dissectionSize);
		dissectionSize = 0;
		packetDissection->setMaximumSize(s);
	}
	else {
		dissectionSize = s.height();
		s.setHeight(0);
		packetDissection->setMaximumSize(s);
	}
}
void Zethereal::expandDump()
{
	if (! ensureNotLast(zoomDump)) return;
	QSize s = packetDump->maximumSize();
	if (dumpSize) {
		s.setHeight(dumpSize);
		dumpSize = 0;
		packetDump->setMaximumSize(s);
	}
	else {
		dumpSize = s.height();
		s.setHeight(0);
		packetDump->setMaximumSize(s);
	}
}



void Zethereal::add_packet(char *time, char *src, char *dst, char *proto, char *info, metapkt_t *mpkt)
{
	char nos[10];
	
	snprintf(nos, 10, "%3i", ++myno);

	new QListViewItem(packetList, QString(nos), QString(time),
			QString(src), QString(dst),
			QString(proto), QString(info));
	mpktvec.push_back(mpkt);
}


char chfilter(char x)
{
	if ((x < 32) || (x>= 127))
		return '.';
	else
		return x;
}
			

void recurse_tree(GNode *node, QWidget *parent, int root)
{
	GNode *n = node;

	/* Qt QListView sorting policies are either alphabetically related
	 * or "last in first displayed". What about "last-in last displayed ?"
	 * Because of that, I need to parse the GNode tree in reverse,
	 * which, it seems, was not expected as a GNode use. 
	 * Hence this hack.
	 */
	
	if (n) {
		while (n->next) { 
			n->next->prev = n;
			n = n->next;
		}
	}

	for (; n; n = n->prev) {
	
		QListViewItem *child;
		char *txt = "??";



		if (n->data)
			txt = (char *)n->data;
			
		if (root) 
			child = new QListViewItem((QListView *)parent, 
					QString(txt));
		else
			child = new QListViewItem((QListViewItem *)parent, 
					QString(txt));

		recurse_tree(n->children, (QWidget *)child, 0);
	}
}
	
	


void Zethereal::dissect_packet(QListViewItem *p)
{
	if (p) {
		int npkt = atoi(p->text(0));


		metapkt_t *mpkt = mpktvec[npkt-1];

		GNode *tree = (GNode *)mpkt->tree;

		unsigned char *pkt = mpkt->pkt;
		
		int sz = mpkt->phdr->caplen;

		int i,j;
		char buf[8192];
		char tmp[10];
		char tmp2[18];

		buf[0]=0;
		tmp2[16]='\n';
		tmp2[17]='\0';
		

		for (i=0; i<sz/16; i++) {
			snprintf(tmp,10,"%.2hx ",i);
			strcat(buf, tmp);
			for(j=0; j<16; j++) {
				snprintf(tmp,10,"%.2x",pkt[i*16+j]);
				strcat(buf, tmp);
				tmp2[j] = chfilter(pkt[i*16+j]);
				if (j%4 == 3) strcat(buf," ");
			}
			strncat(buf, tmp2, 8191);
		}
		buf[8191]=0;
		if (sz%16) {
			snprintf(tmp,10,"%.3hx ",i*16);
			strcat(buf, tmp);
			for(j=0; j<16; j++) {
				if (j < sz%16) 
					snprintf(tmp,10,"%.2x",pkt[i*16+j]);
				else 	
					snprintf(tmp,10,"  ");
				strcat(buf, tmp)	;
				tmp2[j] = chfilter(pkt[i*16+j]);
				if (j%4 == 3) strcat(buf," ");
			}
			tmp2[sz%16] = '\0';
			strcat(buf, tmp2);
		}
		
				
		packetDump->setText(QString(buf));
	
		packetDissection->clear();
		recurse_tree(tree, packetDissection, 1);
		
	}
}


void Zethereal::open(void)
{
	update_engine();
	clearCapture();
	engine->do_read_file();
	mainUI->showPage(dissectTab);
}

void Zethereal::capture(void)
{
	startButton->setEnabled( FALSE );
	openButton->setEnabled( FALSE );
	stopButton->setEnabled( TRUE );	
	update_engine();
	clearCapture();
	engine->do_capture();
	if (engine->is_main_instance()) {
		stopButton->setEnabled( FALSE );	
		startButton->setEnabled( TRUE );
		openButton->setEnabled( TRUE );
	}

}

void Zethereal::stopCapture(void)
{
	stopButton->setEnabled( FALSE );	
	startButton->setEnabled( TRUE );
	openButton->setEnabled( TRUE );
	engine->do_stop_capture();
}

void Zethereal::debug(void)
{
	update_engine();
	engine->dump_state();
}
